" Vim9 legacy script header


" comment

if !has('vim9script')
  # 42 " comment
  source foo.vim
  finish
endif

" comment

vim9script noclear

# comment

 # string only recognised with leading char
 "useless string"

