" The Java source file uses long lines with tabs
" so use a smaller tabstop value, so that no
" soft-wrapping occurs (which may cause the syntax
" tests to miss some lines during scrolling)
setl ts=4
