" Test file for gettext() package makefile
" Last Change:	2024 Jun 01

echo gettext("This is a test", "test_gettext")
echo gettext("This is another test", "test_gettext")
echo gettext("This is a test", "test_gettext")
" vim: ts=8
