# Built-in Constants
# https://docs.python.org/3/library/constants.html

# VIM_TEST_SETUP let g:python_constant_highlight = 1
# VIM_TEST_SETUP let g:python_no_builtin_highlight = 1
test = False
test = True
test = None
test = NotImplemented
test = Ellipsis
test = __debug__
