" Vim Keymap file for Serbian characters, classical variant, UTF-8 encoding
"
" Maintainer:   Aleksandar Veselinovic <aleksa@cs.cmu.edu>
" URL:		    http://www.cs.cmu.edu/~aleksa/personal/vim/keymap/serbian-latin-UTF-8.vim
" Last Changed: 2002 Mar 29

scriptencoding UTF-8

let b:keymap_name = "serbian-UTF-8"

loadkeymap
Y	Z
{	Š
}	Đ
y	z
[	š
]	đ
:   Č
\"	Ć
\|	Ž
;	č
'	ć
\\	ž
"Z	Y
"z	y
<	;
>	:
/	-
_	?
,,	„
''	”

E$	<char-0x20AC>	" EURO SIGN
--	<char-0x2013>	" endash
---	<char-0x2014>	" emdash

" @	"
" #	'
" $	*
" %	:
" ^	,
" &amp;	.
" *	;

