#!/bin/bash
basename
cat
chgrp
chmod
chown
cksum
cmp
comm
cp
cut
date
dirname
egrep		# Obsolescent
expr
fgrep		# Obsolescent
fmt
fold
getconf
grep
head
iconv
id
join
ln
logname
ls
md5sum
mkdir
mkfifo
mktemp
mv
od
paste
pathchk
readlink
realpath
rev
rm
rmdir
sha1sum
sha256sum
sha224sum
sha384sum
sha512sum
stty
sum
sync
tail
tee
tr
tty
uname
uniq
wc
xargs
# xgrep is neither a builtin for bash nor an external program
xgrep
